class T2Disassembler
	attr_accessor :c_parser
	def decompile(entry)
		@c_parser = C::Parser.new
		@c_parser.parse <<EOS
void md5(char*, int, int);
char *malloc(int);
void free(char*);
void decryptcopy(char *dst, char *src);
EOS
		decompile_func(entry)
		@c_parser
	end

	# decompiles a function
	# decompiles subfunction as needed
	def decompile_func(entry)
		entry = normalize entry
		return if not @decoded[entry]

		# create a new toplevel function to hold our code
		func = C::Variable.new
		func.name = auto_label_at(entry, 'func')
		func.type = C::Function.new C::BaseType.new(:void)
		if @c_parser.toplevel.symbol[func.name]
			if not @c_parser.toplevel.statements.grep(C::Declaration).find { |decl| decl.var.name == func.name }
				# recursive dependency: declare prototype
				@c_parser.toplevel.statements << C::Declaration.new(func)
			end
			return
		end
		@c_parser.toplevel.symbol[func.name] = func

		# find decodedinstruction blocks constituing the function
		myblocks = decompile_func_listblocks(entry)

		# rename [ebp+42] => var42
		decompile_makestackvars myblocks.map { |b, to| @decoded[b].block }

		# find stack variable dependencies
		deps = decompile_func_finddeps(myblocks)

		scope = func.initializer = C::Block.new(@c_parser.toplevel)
		# di blocks => raw c statements, declare variables
		stmts = decompile_blocks(myblocks, deps, scope)
		scope.statements = scope.statements.sort_by { |sm| sm.var.name[3..-1].to_i 16 }
		func.type.args = []
		scope.statements.dup.each { |sm|
			if sm.var.name =~ /arg(\d+)/
				func.type.args << scope.statements.delete(sm).var
			end
		}
		func.type.args = func.type.args.sort_by { |sm| -sm.name[/arg(\d+)/, 1].to_i }
		scope.statements.concat stmts

		# goto bla ; bla: goto blo => goto blo ;; goto bla ; bla: return => return ; remove last return if return void
		# (handles loops)
		decompile_simplify_goto(scope)

		# change if() goto to if, if/else, while
		decompile_match_controlseq(scope)

		decompile_remove_labels(scope)

		case ret = scope.statements.last
		when C::CExpression: puts "no return at end of func"
		when C::Return
			if not ret.value
				scope.statements.pop
			else
				func.type.type = C::BaseType.new(:int)
			end
		end

		decompile_c_types(scope)

		@c_parser.toplevel.statements << C::Declaration.new(func)
	end

	# return an array of address of block start, decompile subfunctions
	def decompile_func_listblocks(entry)
		blocks = []
		todo = [entry]
		while a = todo.pop
			next if blocks.find { |aa, at| aa == a }
			next if not di = @decoded[a]
			next if not di.kind_of? DecodedInstruction
			next if di.block.list.first.instruction.opname == 'syscall_free_exitvm'
			blocks << [a, []]
			next if di.block.list.last.instruction.opname == 'ret'
			di.block.each_to { |ta, type|
				next if type == :indirect
				ta = normalize ta
				if @function[ta]	# and di.block.to_subfuncret # XXX __attribute__((noreturn)) ?
					decompile_func(ta) if ta != entry
				else
					auto_label_at(ta, 'label') if blocks.find { |aa, at| aa == ta }
					blocks.last[1] |= [ta]
					todo << ta
				end
			}
		end
		blocks
	end

	# patches instruction arguments to replace [ebp+X] => :varX
	def decompile_makestackvars(blocks)
		tovar = proc { |e|
			case e
			when Expression
				r = tovar[e.rexpr]
				if r.kind_of? ::Symbol and e.op == :+ and not e.lexpr
					r
				else
					Expression[tovar[e.lexpr], e.op, r]
				end
			when Indirection
				if o = Expression[e.target, :-, :ebp].reduce and o.kind_of? ::Integer
					o >= 0 ? "var#{o.to_s 16}".to_sym : "arg#{-4-o}".to_sym
				else Indirection[tovar[e.target], e.len]
				end
			else e
			end
		}
		blocks.each { |block| block.list.each { |di| di.instruction.args.map! { |arg| tovar[arg] } } }
	end

	# turns an Expression to an CExpression, create+declares needed variables in scope
	def decompile_cexpr(e, scope)
		case e
		when Expression
			if e.op == :'=' and e.rexpr.kind_of? Expression and e.rexpr.op == :+ and e.rexpr.lexpr == e.lexpr
				e = Expression[e.lexpr, :'+=', e.rexpr.rexpr]
			end
			if e.lexpr
				a = decompile_cexpr(e.lexpr, scope)
				C::CExpression.new(a, e.op, decompile_cexpr(e.rexpr, scope), a.type)
			elsif e.op == :+
				decompile_cexpr(e.rexpr, scope)
			else
				a = decompile_cexpr(e.rexpr, scope)
				C::CExpression.new(nil, e.op, a, a.type)
			end
		when Indirection
			# XXX int *p ; p + 4*z  =>  p + z
			p = decompile_cexpr(e.target, scope)
			p = C::CExpression.new(nil, nil, p, C::Pointer.new(C::BaseType.new(e.len == 4 ? :int : :char)))
			p = C::CExpression.new(nil, nil, p, p.type) if not p.rexpr.kind_of? C::CExpression
			C::CExpression.new(nil, :*, p, p.type.type)
		when ::Symbol
			name = e.to_s
			if not s = scope.symbol[name]
				s = C::Variable.new
				s.type = C::BaseType.new(:int)
				s.storage = :register if name[0] == ?r
				s.name = name
				scope.statements << C::Declaration.new(scope.symbol[name] = s)
			end
			s
		when ::Integer
			C::CExpression.new(nil, nil, e, C::BaseType.new(:int))
		when ::String
			C::CExpression.new(nil, nil, (e[0] == ?" ? eval(e) : e), C::Pointer.new(C::BaseType.new(:char)))
		when C::CExpression: e
		else puts "decompile_cexpr unhandled #{e.inspect}" ; C::CExpression.new(nil, nil, e, C::BaseType.new(:void))
		end
	end

	# list variable dependency for each block, remove useless writes
	# returns { blockaddr => [list of vars that must be written] }
	def decompile_func_finddeps(blocks)
		deps_r = {} ; deps_w = {} ; deps_to = {}
		blocks.each { |b, to|
			deps_r[b] = [] ; deps_w[b] = [] ; deps_to[b] = to

			@decoded[b].block.list.each { |di|
				a = di.instruction.args.dup
				case di.instruction.opname
				when 'jrz', 'call': a.shift
				when 'ret': a = [:retval]
				when 'mov', 'syscall_alloc_ptr': w = a.shift
				when 'add', 'xor', 'and', 'mul', 'shr', 'xorb1', 'setgt', 'setlt', 'seteql', 'host_call': w = a.first
				when 'syscall_free', 'syscall_md5', 'host_push', 'decryptcopy'
				else puts "block_vardep: unhandled #{di}"
				end
				deps_r[b] |= a.map { |ee| Expression[ee].externals.grep(::Symbol) }.flatten - deps_w[b]
				deps_w[b] |= Expression[w].externals.grep(::Symbol) if w
			}
		}
		deps_w.each { |b, deps|
			deps.delete_if { |dep|
				next if dep == :retval
				ret = true
				done = []
				todo = deps_to[b].dup
				while a = todo.pop
					next if done.include? a
					done << a
					if not deps_r[a] or deps_r[a].include? dep
						ret = false
						break
					elsif not deps_w[a].include? dep
						todo.concat deps_to[a]
					end
				end
				ret
			}
		}
		deps_w
	end

	def decompile_blocks(myblocks, deps, scope, nextaddr = nil)
		stmts = []
		until myblocks.empty?
			b, to = myblocks.shift
			if l = @prog_binding.index(b)
				stmts << C::Label.new(l)
			end

			# list of assignments [[dest reg, expr assigned]]
			ops = []
			# reg binding (reg => value, values.externals = regs at block start)
			binding = {}
			# Expr => CExpr
			ce  = proc { |*e|
				e = Expression[Expression[*e].reduce]
				decompile_cexpr(e, scope)
			}
			# Expr => Expr.bound(binding) => CExpr
			ceb = proc { |*e| ce[Expression[*e].bind(binding).reduce] }
			# shortcut to global funcname => Var (malloc, md5..)
			ts = @c_parser.toplevel.symbol
			# list of exprs that are host_push()ed (for host_call)
			hostcallargs = []

			# dumps a CExprs that implements an assignment to a reg (uses ops[], patches op => [reg, nil])
			commit = proc {
				#ops.each { |r, v| stmts << ce[r, :'=', v] }
				#binding = {}
				deps[b].map { |k|
					[k, ops.rindex(ops.reverse.find { |r, v| r == k })]
				}.sort_by { |k, i| i.to_i }.each { |k, i|
					next if not i or not binding[k]
					e = k
					final = []
					ops[0..i].reverse_each { |r, v|
						final << r if not v
						e = Expression[e].bind(r => v).reduce if not final.include? r
					}
					ops[i][1] = nil
					binding.delete k
					stmts << ce[k, :'=', e]
				}
			}

			# go
			@decoded[b].block.list.each { |di|
				a = di.instruction.args
				case op = di.instruction.opname
				when 'jrz'
					commit[]
					# trivial if / then / end recognition, while
					# should check for while, else, multiblock then...
					stmts << C::If.new(ceb[:'!', a.last], C::Goto.new(a.first))
					to.delete normalize(a.first)
				when 'ret'
					commit[]
					rv = ce[:retval] if deps[b].include? :retval
					stmts << C::Return.new(rv)
				when 'call'
					commit[]
					rt = ts[a.first].type.type
					fc = C::CExpression.new(ts[a.first], :funcall, a[1..-1].map { |e| ceb[e] }, rt)
					if not rt.kind_of? C::BaseType or rt.name != :void
						fc = ce[:retval, :'=', fc]
					end
					stmts << fc
				when 'mov', 'add', 'xor', 'and', 'mul', 'shr', 'xorb1', 'setgt', 'setlt', 'seteql'
					next if op == 'add' and a.first == :esp and o = Expression[a.last].reduce and o.kind_of? ::Integer and o.abs!=4
					ret = case op
					when 'mov'
						Expression[a.last]
					when 'xorb1'
						Expression[[a.first, :&, 1], :^, 1]
					else
						Expression[a.first, 
							{'add'=>:+, 'xor'=>:^, 'and'=>:&, 'mul'=>:*, 'shr'=>:>>, 'setgt'=>:>, 'setlt'=>:<, 'seteql'=>:==}[op],
							a.last]
					end

					if a.first.kind_of? ::Symbol
						ops << [a.first, ret]
						binding[a.first] = Expression[ret.bind(binding).reduce]
					else
						stmts << ceb[a.first, :'=', ret]
					end
				when 'syscall_alloc_ptr'
					p = C::Pointer.new(C::BaseType.new(:char))
					stmts << C::CExpression.new(ce[a.first], :'=', C::CExpression.new(ts['malloc'], :funcall, [ceb[a.last]], p), p)
					binding.delete a.first
				when 'syscall_free'
					stmts << C::CExpression.new(ts['free'], :funcall, [ceb[a.last]], C::BaseType.new(:void))
				when 'syscall_md5'
					stmts << C::CExpression.new(ts['md5'], :funcall, a.map { |e| ceb[e] }, C::BaseType.new(:void))
				when 'decryptcopy'
					stmts << C::CExpression.new(ts['decryptcopy'], :funcall, a.map { |e| ceb[e] }, C::BaseType.new(:void))
				when 'host_push'
					hostcallargs.unshift ceb[a.first]
				when 'host_call'
					fptr = ceb[a.first]
					proto = C::Function.new(C::BaseType.new(:int))
					proto.args = [C::BaseType.new(:int)] * hostcallargs.length
					fptr = C::CExpression.new(nil, nil, fptr, C::Pointer.new(proto)) if not fptr.kind_of? C::CExpression
					fptr = C::CExpression.new(nil, nil, fptr, C::Pointer.new(proto))
					stmts << ce[a.first, :'=', C::CExpression.new(fptr, :funcall, hostcallargs, proto.type)]
					binding.delete a.first
					hostcallargs = []
				else puts "unknown instr to decompile: #{di}"
				end
			}
			commit[]
			puts "  block #{b} ends with nonempty hostcallargs" if not hostcallargs.empty?
			case to.length
			when 0
				if not myblocks.empty? and @decoded[b].block.list.last.instruction.opname != 'ret'
					puts "  block #{b} has no to and don't end in ret"
				end
			when 1
				if (myblocks.empty? ? nextaddr != to[0] : myblocks.first.first != to[0])
					stmts << C::Goto.new(auto_label_at(to[0], 'unknown_goto'))
				end
			else
				puts "  block #{b} with multiple to"
			end
		end
		stmts
	end

	# simplify goto -> goto
	# iterative process, to not infinite loop on b:goto a; a:goto b;
	# TODO multipass ? (goto a -> goto b -> goto c -> goto d)
	# remove last return if not useful
	def decompile_simplify_goto(scope)
		cntr = -1
		simpler_goto = proc { |g|
			case ret = g
			when C::Goto
				# return a new goto
				decompile_walk(scope) { |s|
					if s.kind_of? C::Block and l = s.statements.grep(C::Label).find { |l| l.name == g.target }
						case nt = s.statements[s.statements.index(l)..-1].find { |ss| not ss.kind_of? C::Label }
						when C::Goto: ret = nt
						end
					end
				}
			when C::Return
				# XXX if () { return } else { return }
				lr = scope.statements.last
				if g != lr and lr.kind_of? C::Return and g.value == lr.value
					if not scope.statements[-2].kind_of? C::Label
						scope.statements.insert(-2, C::Label.new("ret_#{cntr += 1}", nil))
					end
					ret = C::Goto.new(scope.statements[-2].name)
				end
			end
			ret
		}
		decompile_walk(scope) { |s|
			case s
			when C::Block
				s.statements.each { |ss|
					s.statements[s.statements.index(ss)] = simpler_goto[ss]
				}
			when C::If
				s.bthen = simpler_goto[s.bthen]
			end
		}
	end

	def decompile_remove_labels(scope)
		decompile_walk(scope) { |s|
			next if not s.kind_of? C::Block
			s.statements.delete_if { |l|
				if l.kind_of? C::Label
					notfound = true
					decompile_walk(scope) { |ss| notfound = false if ss.kind_of? C::Goto and ss.target == l.name }
				end
				notfound
			}
		}
	end

	# changes ifgoto, goto to while/ifelse..
	def decompile_match_controlseq(scope)
		scope.statements = decompile_cseq_if(scope.statements, scope)
		decompile_cseq_while(scope.statements)
	end

	# ifgoto => ifthen
	# ary is an array of statements where we try to find if () {} [else {}]
	# recurses to then/else content
	def decompile_cseq_if(ary, scope)
		# helper to negate the if condition
		negate = proc { |ce|
			if ce.kind_of? C::CExpression and nop = { :== => :'!=', :'!=' => :==, :> => :<=, :>= => :<, :< => :>=, :<= => :>, :'!' => :'!' }[ce.op]
				if nop == :'!'
					ce.rexpr
				else
					C::CExpression.new(ce.lexpr, nop, ce.rexpr, ce.type)
				end
			else
				C::CExpression.new(nil, :'!', ce, C::BaseType.new(:int))
			end
		}

		# the array of decompiled statements to use as replacement
		ret = []
		# list of labels appearing in ary
		inner_labels = ary.grep(C::Label).map { |l| l.name }
		while s = ary.shift
			# "forward" ifs only
			if s.kind_of? C::If and s.bthen.kind_of? C::Goto and l = ary.grep(C::Label).find { |l| l.name == s.bthen.target }
				# if {goto l;} a; l: => if (!) {a;}
				s.test = negate[s.test]
				s.bthen = C::Block.new(scope)
				s.bthen.statements = decompile_cseq_if(ary[0...ary.index(l)], scope)
				bts = s.bthen.statements
				ary[0...ary.index(l)] = []

				# if { a; goto outer; } b; return; => if (!) { b; return; } a; goto outer;
				if bts.last.kind_of? C::Goto and not inner_labels.include? bts.last.target and g = ary.find { |ss| ss.kind_of? C::Goto or ss.kind_of? C::Return } and g.kind_of? C::Return
					s.test = negate[s.test]
					ary[0..ary.index(g)], bts[0..-1] = bts, ary[0..ary.index(g)]
				end

				# if { a; goto l; } b; l: => if {a;} else {b;}
				if bts.last.kind_of? C::Goto and l = ary.grep(C::Label).find { |l| l.name == bts.last.target }
					s.belse = C::Block.new(scope)
					s.belse.statements = decompile_cseq_if(ary[0...ary.index(l)], scope)
					ary[0...ary.index(l)] = []
					bts.pop
				end

				# if { a; l: b; goto any;} c; goto l; => if { a; } else { c; } b; goto any;
				if not s.belse and (bts.last.kind_of? C::Goto or bts.last.kind_of? C::Return) and g = ary.grep(C::Goto).first and l = bts.grep(C::Label).find { |l| l.name == g.target }
					s.belse = C::Block.new(scope)
					s.belse.statements = decompile_cseq_if(ary[0...ary.index(g)], scope)
					ary[0..ary.index(g)], bts[bts.index(l)..-1] = bts[bts.index(l)..-1], []
				end

				# if { a; b; c; } else { d; b; c; } => if {a;} else {d;} b; c;
				if s.belse
					bes = s.belse.statements
					while not bts.empty?
						if bts.last.kind_of? C::Label: ary.unshift bts.pop
						elsif bes.last.kind_of? C::Label: ary.unshift bes.pop
						elsif bts.last.to_s == bes.last.to_s: ary.unshift bes.pop ; bts.pop
						else break
						end
					end

					# if () { a; } else { b; } => if () { a; } else b;
					# if () { a; } else {} => if () { a; }
					case bes.length
					when 0: s.belse = nil
					when 1: s.belse = bes.first
					end
				end

				# if () {} else { a; } => if (!) { a; }
				# if () { a; } => if () a;
				case bts.length
				when 0: s.test, s.bthen, s.belse = negate[s.test], s.belse, nil if s.belse
				when 1: s.bthen = bts.first
				end
			end
			ret << s
		end
		ret
	end

	def decompile_cseq_while(ary)
		# find the next instruction that is not a label
		ni = proc { |l| ary[ary.index(l)..-1].find { |s| not s.kind_of? C::Label } }
		ary.each { |s|
			case s
			when C::Label
				if ss = ni[s] and ss.kind_of? C::If and not ss.belse and ss.bthen.kind_of? C::Block and ss.bthen.statements.last.kind_of? C::Goto and ss.bthen.statements.last.target == s.name
					ss.bthen.statements.pop
					if l = ary[ary.index(ss)+1] and l.kind_of? C::Label
						ss.bthen.statements.grep(C::If).each { |i|
							i.bthen = C::Break.new if i.bthen.kind_of? C::Goto and i.bthen.target == l.name
						}
					end
					ary[ary.index(ss)] = C::While.new(ss.test, ss.bthen)
				end
			when C::If
				decompile_cseq_while(s.bthen.statements) if s.bthen.kind_of? C::Block
				decompile_cseq_while(s.belse.statements) if s.belse.kind_of? C::Block
			when C::While
				decompile_cseq_while(s.body.statements) if s.body.kind_of? C::Block
			end
		}
	end

	def decompile_walk_ce(ce, &b)
		case ce
		when C::CExpression
			yield ce
			decompile_walk_ce(ce.lexpr, &b)
			decompile_walk_ce(ce.rexpr, &b)
		when ::Array
			ce.each { |ce| decompile_walk(ce, &b) }
		end
	end

	def decompile_c_types(scope)
		# pass 1: find variables type
		types = {}
		# scan for var = int
		decompile_walk(scope) { |ce| decompile_walk_ce(ce) { |ce|
			if ce.op == :'=' and ce.rexpr.kind_of? C::CExpression and ce.rexpr.op == nil and ce.rexpr.rexpr.kind_of? ::Integer
				v = ce.lexpr
				if v.kind_of? C::Variable and not v.storage
					types[v.name] ||= ce.rexpr.type
				end
			elsif ce.op == :'=' and ce.rexpr.kind_of? C::Variable and types[ce.rexpr.name]
				v = ce.lexpr
				if v.kind_of? C::Variable and not v.storage
					types[v.name] ||= types[ce.rexpr.name]
				end
			end
		} }
		# scan for *(bla*)x
		decompile_walk(scope) { |ce| decompile_walk_ce(ce) { |ce|
			if ce.op == :* and not ce.lexpr and ce.rexpr.kind_of? C::CExpression and ce.rexpr.op == nil and ce.rexpr.rexpr.kind_of? C::CExpression
				p = ce.rexpr.rexpr
				if p.op == nil and p.rexpr.kind_of? C::Variable and not p.rexpr.storage
					types[p.rexpr.name] ||= ce.rexpr.type
				elsif p.op == :+ and p.lexpr
					if p.lexpr.kind_of? C::Variable and not p.lexpr.storage and (p.rexpr.kind_of? C::Variable and types[p.rexpr.name] and types[p.rexpr.name].integral?) or (p.rexpr.kind_of? C::CExpression and not p.rexpr.op and p.rexpr.rexpr.kind_of? ::Integer)
						types[p.lexpr.name] ||= ce.rexpr.type
					elsif p.rexpr.kind_of? C::Variable and not p.rexpr.storage and p.lexpr.kind_of? C::Variable and types[p.lexpr.name] and types[p.lexpr.name].integral?
						types[p.rexpr.name] ||= ce.rexpr.type
					end
				end
			end
		} }

		types.each { |k, v| scope.symbol[k].type = v }
		# pass 2: fix indirections & pointer addition
		# TODO (char)toto == 42 => toto == 'A'
		decompile_walk(scope) { |ce| decompile_walk_ce(ce) { |ce|
			# *(bla*)x => *x, *(bla*)(x+2) => x[2/sizeof(*x)]
			if ce.op == :* and not ce.lexpr and ce.rexpr.kind_of? C::CExpression and ce.rexpr.op == nil and ce.rexpr.rexpr.kind_of? C::CExpression
				p = ce.rexpr.rexpr
				if p.op == nil and p.rexpr.kind_of? C::Variable and not p.rexpr.storage and types[p.rexpr.name] == ce.rexpr.type
					#ce.rexpr = ce.rexpr.rexpr
					ce.op = :[]
					ce.lexpr = ce.rexpr.rexpr
					ce.rexpr = C::CExpression.new(nil, nil, 0, C::BaseType.new(:int))
				elsif p.op == :+ and p.lexpr
					if p.lexpr.kind_of? C::Variable and types[p.lexpr.name] == ce.rexpr.type
						ptr = p.lexpr
						idx = p.rexpr
					elsif p.rexpr.kind_of? C::Variable and types[p.rexpr.name] == ce.rexpr.type
						ptr = p.rexpr
						idx = p.lexpr
					else next
					end
					ce.op = :[]
					ce.lexpr = ptr
					if idx.kind_of? C::CExpression and not idx.op and idx.rexpr.kind_of? ::Integer
						ce.rexpr = idx
						idx.rexpr /= @c_parser.sizeof(ce)
					else
						sz = @c_parser.sizeof(ce)
						ce.rexpr = (sz == 1 ? idx : C::CExpression.new(idx, :/, C::CExpression.new(nil, nil, sz, idx.type), idx.type))
					end
				end
			elsif ce.op == :'+=' and ce.lexpr.kind_of? C::Variable and types[ce.lexpr.name] and types[ce.lexpr.name].kind_of? C::Pointer
				if ce.rexpr.kind_of? C::CExpression and not ce.rexpr.op and ce.rexpr.rexpr.kind_of? ::Integer
					ce.rexpr.rexpr /= @c_parser.sizeof(ce)
				else
					sz = @c_parser.sizeof(ce)
					ce.rexpr = C::CExpression.new(ce.rexpr, :/, C::CExpression.new(nil, nil, sz, ce.rexpr.type), ce.rexpr.type) if sz != 1
				end
			end
		} }
	end

	# yields each statement (recursive)
	def decompile_walk(scope, &b)
		case scope
		when ::Array: scope.each { |s| decompile_walk(s, &b) }
		when C::Statement
			yield scope
			case scope
			when C::Block: decompile_walk(scope.statements, &b)
			when C::If
				yield scope.test
				decompile_walk(scope.bthen, &b)
				decompile_walk(scope.belse, &b) if scope.belse
			when C::While
				yield scope.test
				decompile_walk(scope.body, &b)
			end
		end
	end
end
