#!/usr/bin/ruby

# this metasm scripts compiles an ELF that will solve the T2 2007 challenge

require 'metasm'
Metasm::ELF.assemble(Metasm::Ia32.new, DATA.read).encode_file('resolv')

__END__

.pt_gnu_stack rw
.interp nil

.entrypoint
call bruteforce
call enumerate

// exit
mov eax, 1
mov ebx, 0
int 80h
ret


// **************** h1 bruteforce ****************
bruteforce:
// md5 padding for 16byte buffer
mov ecx, (64-16)/4
mov edi, buf+16
xor eax, eax
rep stosd
mov byte ptr [buf+16], 0x80
mov byte ptr [buf+64-8], 16*8

// bf 1<<20->1
mov ebp, (1<<20)+1
bfh1_loop:
dec ebp
jz bfh1_done

// decrypt
mov esi, ciphered
mov ecx, ciphered+13
mov edi, buf
mov ebx, ebp
decryptloop:
lodsb
xor al, bl
stosb
shr ebx, 1
cmp esi, ecx
jnz decryptloop

// md5sum buf
mov esi, buf
call md5_round

// check md5
cmp eax, 0x748ff717
jnz bfh1_loop
cmp ebx, 0x156aff74
jnz bfh1_loop
cmp ecx, 0xaa8141be
jnz bfh1_loop
cmp edx, 0x30989474
jnz bfh1_loop

// save value
mov dword ptr [h1], ebp
and dword ptr [h1], 0xfffff	// just in case 0x100000 was a match

// puts decryptbuf[0, 16] + ' ' + bp.to_s 16
mov edi, buf+16
mov al, ' '
stosb

mov ebx, hexmap
mov ecx, 8*4
hexloop:
sub ecx, 4
mov eax, ebp
shr eax, cl
and eax, 0xf
xlat
stosb
test ecx, ecx
jnz hexloop

mov al, '\n'
stosb

mov eax, 4	// sys_write
mov ebx, 1	// stdout
mov ecx, buf
mov edx, 16+8+2	// strlen
int 80h

// restore md5 padding
mov ecx, 48/4
mov edi, buf+16
xor eax, eax
rep stosd
mov byte ptr [buf+16], 0x80
mov byte ptr [buf+64-8], 16*8

jmp bfh1_loop


bfh1_done:
ret

// **************** enumeration ****************
enumerate:

// try all possible combinaison, filter out '+' and '/'
// md5 padding for 8bytes
mov ecx, (64-8)/4
mov edi, buf+8
xor eax, eax
rep stosd
mov byte ptr [buf+8], 0x80
mov byte ptr [buf+64-8], 8*8

// final "\n"
mov byte ptr [b64str+16], '\n'

// counter: low 16bits for h2, high 12bits for h1
mov ebp, (1<<28)+1
bfh2_loop:
dec ebp
jz bfh2_done

// h1
mov eax, ebp
shr eax, 16
shl eax, 20
or eax, [h1]
mov [buf], eax

// h2
mov eax, ebp
xor eax, 0x5432
shl eax, 16
mov ax, bp
mov [buf+4], eax

// h3
mov esi, buf
call md5_round
xor eax, ebx
xor eax, ecx
xor eax, edx
mov [b64_3], eax	// no bswap!

// prepare b64 decoding
mov eax, [buf+4]
xor eax, [b64_3]
mov [b64_2], eax

mov eax, [buf]
xor eax, [b64_3]
mov [b64_1], eax

// b64decode (bigendian: decode 4*3bytes)
mov ebx, b64map
mov esi, b64_1
mov edi, b64str
b64outer:
mov edx, [esi]
add esi, 3
bswap edx

mov ecx, 4
b64inner:
rol edx, 6
shl eax, 8
mov al, dl
and al, 0x3f
cmp al, 0x3e
jae bfh2_loop	// badchar !
xlat
loop b64inner
bswap eax
stosd

cmp esi, b64_1+12
jnz b64outer


// output string
mov eax, 4	// sys_write
mov ebx, 1	// stderr
mov ecx, b64str
mov edx, 17	// strlen
int 80h

jmp bfh2_loop

bfh2_done:
ret


// **************** MD5 ****************

// eax, ebx, ecx, edx: hash
// esi: pointer to md5 buffer (already padded)
md5_round:
// init md5
mov eax, 0x67452301
mov ebx, 0xefcdab89
mov ecx, 0x98badcfe
mov edx, 0x10325476
#define A eax
#define B ebx
#define C ecx
#define D edx
// from Christophe Devine's md5.c
push edi	// tmp storage of F()
#define P(a,b,c,d,k,s,t) F(b, c, d) add a, dword ptr [esi+4*k] lea a, [a + edi + t] rol a, s add a, b
#define F(x,y,z) mov edi, z  xor edi, y  and edi, x  xor edi, z // (z ^ (x & (y ^ z)))
    P( A, B, C, D,  0,  7, 0xD76AA478)
    P( D, A, B, C,  1, 12, 0xE8C7B756)
    P( C, D, A, B,  2, 17, 0x242070DB)
    P( B, C, D, A,  3, 22, 0xC1BDCEEE)
    P( A, B, C, D,  4,  7, 0xF57C0FAF)
    P( D, A, B, C,  5, 12, 0x4787C62A)
    P( C, D, A, B,  6, 17, 0xA8304613)
    P( B, C, D, A,  7, 22, 0xFD469501)
    P( A, B, C, D,  8,  7, 0x698098D8)
    P( D, A, B, C,  9, 12, 0x8B44F7AF)
    P( C, D, A, B, 10, 17, 0xFFFF5BB1)
    P( B, C, D, A, 11, 22, 0x895CD7BE)
    P( A, B, C, D, 12,  7, 0x6B901122)
    P( D, A, B, C, 13, 12, 0xFD987193)
    P( C, D, A, B, 14, 17, 0xA679438E)
    P( B, C, D, A, 15, 22, 0x49B40821)
#undef F
#define F(x,y,z) mov edi, x  xor edi, y  and edi, z  xor edi, y //(y ^ (z & (x ^ y)))
    P( A, B, C, D,  1,  5, 0xF61E2562)
    P( D, A, B, C,  6,  9, 0xC040B340)
    P( C, D, A, B, 11, 14, 0x265E5A51)
    P( B, C, D, A,  0, 20, 0xE9B6C7AA)
    P( A, B, C, D,  5,  5, 0xD62F105D)
    P( D, A, B, C, 10,  9, 0x02441453)
    P( C, D, A, B, 15, 14, 0xD8A1E681)
    P( B, C, D, A,  4, 20, 0xE7D3FBC8)
    P( A, B, C, D,  9,  5, 0x21E1CDE6)
    P( D, A, B, C, 14,  9, 0xC33707D6)
    P( C, D, A, B,  3, 14, 0xF4D50D87)
    P( B, C, D, A,  8, 20, 0x455A14ED)
    P( A, B, C, D, 13,  5, 0xA9E3E905)
    P( D, A, B, C,  2,  9, 0xFCEFA3F8)
    P( C, D, A, B,  7, 14, 0x676F02D9)
    P( B, C, D, A, 12, 20, 0x8D2A4C8A)
#undef F
#define F(x,y,z) mov edi, x  xor edi, y  xor edi, z // (x ^ y ^ z)
    P( A, B, C, D,  5,  4, 0xFFFA3942)
    P( D, A, B, C,  8, 11, 0x8771F681)
    P( C, D, A, B, 11, 16, 0x6D9D6122)
    P( B, C, D, A, 14, 23, 0xFDE5380C)
    P( A, B, C, D,  1,  4, 0xA4BEEA44)
    P( D, A, B, C,  4, 11, 0x4BDECFA9)
    P( C, D, A, B,  7, 16, 0xF6BB4B60)
    P( B, C, D, A, 10, 23, 0xBEBFBC70)
    P( A, B, C, D, 13,  4, 0x289B7EC6)
    P( D, A, B, C,  0, 11, 0xEAA127FA)
    P( C, D, A, B,  3, 16, 0xD4EF3085)
    P( B, C, D, A,  6, 23, 0x04881D05)
    P( A, B, C, D,  9,  4, 0xD9D4D039)
    P( D, A, B, C, 12, 11, 0xE6DB99E5)
    P( C, D, A, B, 15, 16, 0x1FA27CF8)
    P( B, C, D, A,  2, 23, 0xC4AC5665)
#undef F
#define F(x,y,z) mov edi, z  not edi  or edi, x  xor edi, y // (y ^ (x | ~z))
    P( A, B, C, D,  0,  6, 0xF4292244)
    P( D, A, B, C,  7, 10, 0x432AFF97)
    P( C, D, A, B, 14, 15, 0xAB9423A7)
    P( B, C, D, A,  5, 21, 0xFC93A039)
    P( A, B, C, D, 12,  6, 0x655B59C3)
    P( D, A, B, C,  3, 10, 0x8F0CCC92)
    P( C, D, A, B, 10, 15, 0xFFEFF47D)
    P( B, C, D, A,  1, 21, 0x85845DD1)
    P( A, B, C, D,  8,  6, 0x6FA87E4F)
    P( D, A, B, C, 15, 10, 0xFE2CE6E0)
    P( C, D, A, B,  6, 15, 0xA3014314)
    P( B, C, D, A, 13, 21, 0x4E0811A1)
    P( A, B, C, D,  4,  6, 0xF7537E82)
    P( D, A, B, C, 11, 10, 0xBD3AF235)
    P( C, D, A, B,  2, 15, 0x2AD7D2BB)
    P( B, C, D, A,  9, 21, 0xEB86D391)
#undef F
#undef P
pop edi
add eax, 0x67452301
add ebx, 0xefcdab89
add ecx, 0x98badcfe
add edx, 0x10325476
ret

.data
hexmap db "0123456789abcdef"
b64map db "FSECURe20O7NBklm1DcdPAG56HLMnopuQvwxKVWqrstXYZyzabTIJfghij3489+/"
ciphered db "X\244{\022\322\246\023\\|\350\350\201\210"
.align 16
b64:
b64_1 dd 0
b64_2 dd 0
b64_3 dd 0
h1 dd ?
.align 16
buf db 64 dup(?)
b64str db 17 dup(?)
