require 'gtk2'

# A little animation in GTK to show the working of an addition by the virtual machine

class Box
	attr_accessor :text, :fontcolor, :color, :rect
	def initialize(text, fontcolor=:black, color=:white, rect=[0, 0])
		@text = text
		@fontcolor = fontcolor
		@color = color
		@rect = rect
	end
end

class Anim < Gtk::DrawingArea
	BG = :paleblue

	# each frame is described here
	# 1
	k1 = 0x12345678
	k2 = 0x9abcdef0
	nh = 0x2438
	ni = 0x9214
	Frames = []
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08X' % [k2^nh, k1^ni, k1^0x64, k1^0x68], :black, :palegreen, [20, 50]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024, 341], :black, :palegreen, [20, 180]),
		Box.new('addition virtuelle', :black, BG, [150, 280])
	]
	# 2
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08X' % [k2^nh, k1^ni, k1^0x64, k1^0x68], :black, :palegreen, [20, 50]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08X' % [nh, ni, 0x64, 0x68], :darkred, :palegreen, [20, 85]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024, 341], :black, :palegreen, [20, 180]),
		Box.new('dechiffre l\'instruction', :black, BG, [150, 280])
	]
	# 3
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08X' % [k2^nh, k1^ni, k1^0x64, k1^0x68], :black, :palegreen, [20, 50]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08X' % [nh, ni, 0x64, 0x68], :darkred, :palegreen, [20, 85]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024+341, 341], :black, :palegreen, [20, 180]),
		Box.new('realise l\'addition', :black, BG, [150, 280])
	]
	# 4
	k1 = 0x56789012
	k2 = 0x82938485
	nh = 0x4234
	ni = 0x8283
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X' % [k2^nh, k1^ni, k1^0x64], :black, :palegreen, [20, 50]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024+341, 341], :black, :palegreen, [20, 180]),
		Box.new('transition a l\'instruction suivante', :black, BG, [150, 280])
	]
	# 5
	k1 = 0x02938493
	k2 = 0x345ae323
	nh = 0x1322
	ni = 0x4223
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X 0x%08x 0x%08x' % [k2^nh, k1^ni, k1^0x68, k1^0x42, k1^0x23456], :black, :palegreen, [20, 50]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024+341, 0], :black, :palegreen, [20, 180]),
		Box.new('etc', :black, BG, [150, 280])
	]
	# 6
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X' % [0x13e34258, 0x69898384, 0x53132456], :black, :palegreen, [20, 50]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024+341, 2349], :black, :palegreen, [20, 180]),
		Box.new('etc', :black, BG, [150, 280])
	]
	# 7
	Frames << [
		Box.new('ebp', :black, BG, [20, 20]),
		Box.new('0x%08X 0x%08X 0x%08X' % [0x234efa1, 0x3248efa3, 0x9a1e9eef], :black, :palegreen, [20, 50]),
		Box.new('ebx', :black, BG, [20, 150]),
		Box.new('... 0x%08X 0x%08X ...' % [1024+341+2349, 2349], :black, :palegreen, [20, 180]),
		Box.new('etc', :black, BG, [150, 280])
	]

	def initialize
		super()
		set_can_focus true
		@frame_nr = 0

		signal_connect('key_press_event') { |w, ev|
			case ev.keyval
			when Gdk::Keyval::GDK_space, Gdk::Keyval::GDK_Right
				@frame_nr += 1 if @frame_nr < Frames.length - 1
				redraw
			when Gdk::Keyval::GDK_BackSpace, Gdk::Keyval::GDK_Left
				@frame_nr -= 1 if @frame_nr > 0
				redraw
			when Gdk::Keyval::GDK_F4
				Gtk.main_quit
			end
		}

		signal_connect('expose_event') {
			paint
		}

		@color = {}
		signal_connect('realize') {
			{ :white => 'fff', :palegrey => 'ddd', :black => '000', :grey => '444',
			  :red => 'f00', :darkred => '800', :palered => 'fcc',
			  :green => '0f0', :darkgreen => '080', :palegreen => 'cfc',
			  :blue => '00f', :darkblue => '008', :paleblue => 'ccf',
			  :yellow => 'ff0', :darkyellow => '440', :paleyellow => 'ffc',
			}.each { |tag, val|
				@color[tag] = Gdk::Color.new(*val.unpack('CCC').map { |c| (c.chr*4).hex })
			}
			@color.each_value { |c| window.colormap.alloc_color(c, true, true) }
			modify_bg Gtk::STATE_NORMAL, @color[BG]
		}

		@layout = Pango::Layout.new Gdk::Pango.context
		@layout.font_description = Pango::FontDescription.new('courier 16')
	end

	def redraw
		window.invalidate Gdk::Rectangle.new(0, 0, 100000, 1000000), false
	end

	def paint
		w = window
		gc = Gdk::GC.new w
		f = Frames[@frame_nr]

		gc.set_foreground @color[:white]
		w.draw_rectangle(gc, true, @frame_nr*15+5, 5, 10, 10)

		f.each { |box|
			@layout.text = box.text
			box.rect[2] ||= @layout.pixel_size[0]+6
			box.rect[3] ||= @layout.pixel_size[1]+6
			gc.set_foreground @color[box.color]
			w.draw_rectangle(gc, true, *box.rect)
			gc.set_foreground @color[box.fontcolor]
			w.draw_layout(gc, box.rect[0]+3, box.rect[1]+3, @layout)
		}
	end
end

w = Gtk::Window.new
w.add Anim.new
w.set_default_size 800, 500
w.signal_connect('destroy') { Gtk.main_quit }
w.show_all
Gtk.main

